/*
Module : ScreenSaver.cpp
Purpose: Implementation for an MFC class framework to encapsulate screen savers
Created: PJN / 06-05-2000
History: PJN / 06-01-2006 1. Updated the copyright details and general code cleanup.
                          2. CScreenSaverApp::ChangePassword method now returns a BOOL to indicate success.
                          3. Fixed a number of issues when code is compiled using the "Detect 64 bit 
                          portability issues" in Visual Studio .NET 2003. To address this issue, it does
                          mean that you require the Platform SDK to be installed and configured if you 
                          want to compile the code in Visual C++ 6.
                          4. Removed some unnecessary comments from CScreenSaverApp::ShowConfigureDialog.
                          5. Fixed a bug in CScreenSaveWnd::OnClose where it would incorrectly prompt
                          for a password when the screen saver preview was being displayed and you were
                          running on Windows 9x. Thanks to Jason Yang for reporting this bug.
         PJN / 18-07-2006 1. Code now uses new C++ style casts rather than old style C casts where necessary.
                          2. Replaced calls to ZeroMemory with memset
                          3. Reviewed TRACE statements for correctness
                          4. Verified code compiles cleanly on VC 2005
                          5. Addition of CSCREENSAVER_EXT_CLASS preprocessor macro to allow the classes
                          to be easily added to an extension dll.
         PJN / 30-01-2016 1. Updated copyright details.
                          2. Remove VC 6 style classwizard style comments from the code
                          3. Removed the code in the classes which supports Windows 9x.
                          4. Updated the code to compile cleanly in VC 2005 - 2015.
                          5. Updated the sample apps main icon. 
                          6. Added SAL annotations to all the code.
                          7. CScreenSaverApp::InitInstance now calls the base class InitInstance implementation
                          8. Removed unnecessary CScreenSaverWnd::m_bDialogUp member variable
         PJN / 29-10-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean that
                          the code will now only compile on VC 2017 or later.
         PJN / 03-06-2019 1. Updated copyright details.
                          2. Updated the code to clean compile on VC 2019
         PJN / 23-12-2019 1. Fixed various Clang-Tidy static code analysis warnings in the code.
         PJN / 25-03-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 25-04-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations.

Copyright (c) 2000 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/

//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#include "ScreenSaver.h"


//////////////////// Macros / Defines /////////////////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif


//////////////////// Implementation ///////////////////////////////////////////

CScreenSaverCommandLineInfo::CScreenSaverCommandLineInfo() noexcept : m_nShellCommand{ShellCommand::ShowConfigureDialog}, 
                                                                      m_hParentWnd{nullptr}
{
}

void CScreenSaverCommandLineInfo::ParseParam(const TCHAR* pszParam, BOOL /*bFlag*/, BOOL /*bLast*/) noexcept
{
  if (_tcsicmp(pszParam, _T("c")) == 0)
    m_nShellCommand = ShellCommand::ShowConfigureDialog;
  else if (_tcsicmp(pszParam, _T("p")) == 0)
    m_nShellCommand = ShellCommand::PreviewMode;
  else if (_tcsicmp(pszParam, _T("s")) == 0)
    m_nShellCommand = ShellCommand::ScreenSaverMode;
  else if ((_tcsstr(pszParam, _T("c:")) || _tcsstr(pszParam, _T("C:"))) && _tcslen(pszParam) > 2) //Handle the undocumented format of "/c:HWND"
  {
    m_nShellCommand = ShellCommand::ShowConfigureDialog;
  #ifdef _WIN64
#pragma warning(suppress: 26481 26490)
    m_hParentWnd = reinterpret_cast<HWND>(_ttoi64(pszParam + 2));
  #else
#pragma warning(suppress: 26481 26490)
    m_hParentWnd = reinterpret_cast<HWND>(_ttoi(pszParam + 2));
  #endif //#ifdef _WIN64
  }
  else
  {
  //Convert the parameter to a window handle
  #ifdef _WIN64
#pragma warning(suppress: 26490)
    m_hParentWnd = reinterpret_cast<HWND>(_ttoi64(pszParam));
  #else
#pragma warning(suppress: 26490)
    m_hParentWnd = reinterpret_cast<HWND>(_ttoi(pszParam));
  #endif //#ifdef _WIN64
  }
}


#pragma warning(suppress: 26409 26433 26440 26477)
IMPLEMENT_DYNCREATE(CScreenSaverWnd, CWnd)

#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CScreenSaverWnd, CWnd) //NOLINT(modernize-avoid-c-arrays)
  ON_WM_SYSCOMMAND()
  ON_WM_SETCURSOR()
  ON_WM_ACTIVATE()
  ON_WM_ACTIVATEAPP()
  ON_WM_MOUSEMOVE()
  ON_WM_LBUTTONDOWN()
  ON_WM_RBUTTONDOWN()
  ON_WM_MBUTTONDOWN()
  ON_WM_KEYDOWN()
  ON_WM_SYSKEYDOWN()
  ON_WM_PAINT()
  ON_WM_CLOSE()
END_MESSAGE_MAP()

CScreenSaverWnd::CScreenSaverWnd() noexcept : m_bGotCursorPos{false},
                                              m_bPreview{false}
{
}

bool CScreenSaverWnd::GetBoundingRect(_In_ bool bChild, _In_opt_ CWnd* pParentWnd, _Inout_ CRect& rRect)
{
  rRect.SetRectEmpty();

  if (bChild)
  {
    if ((pParentWnd == nullptr) || !pParentWnd->GetSafeHwnd())
    {
      SetLastError(ERROR_INVALID_PARAMETER);
      return false;
    }
    pParentWnd->GetClientRect(&rRect);
  }
  else
  {
    //Get the bounding rect of the screen, This
    //will also handle the case of multiple monitors
    HDC hScreenDC{::GetDC(nullptr)};
    ::GetClipBox(hScreenDC, rRect);
    ::ReleaseDC(nullptr, hScreenDC);
  }

  return true;
}

#pragma warning(suppress: 26440)
DWORD CScreenSaverWnd::GetWindowStyle(_In_ bool bChild, _In_opt_ CWnd* /*pParentWnd*/)
{
  //Calculate the window and extended window styles
  DWORD dwStyle{0};
  if (bChild)
    dwStyle = WS_CHILD | WS_VISIBLE | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
  else
    dwStyle = WS_POPUP | WS_VISIBLE | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;

  return dwStyle;
}

#pragma warning(suppress: 26440)
DWORD CScreenSaverWnd::GetExtendedWindowStyle(_In_ bool bChild, _In_opt_ CWnd* /*pParentWnd*/)
{
#pragma warning(suppress: 26496)
  DWORD dwExStyle{0};
  if (!bChild)
  {
    //For testing purposes, do not make the window topmost
  #ifndef _DEBUG
    dwExStyle = WS_EX_TOPMOST;
  #endif
  }

  return dwExStyle;
}

BOOL CScreenSaverWnd::Create(_In_ bool bChild, _In_opt_ CWnd* pParentWnd)
{
  //Register a class with an empty cursor and a black background
  HCURSOR hCursor{LoadCursor(nullptr, IDC_ARROW)};
  LPCTSTR pszClassName{AfxRegisterWndClass(CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS, hCursor, static_cast<HBRUSH>(GetStockObject(BLACK_BRUSH)))};

  //Get the bounding rect for the window
  CRect rRect;
  if (!GetBoundingRect(bChild, pParentWnd, rRect))
  {
    TRACE(_T("CScreenSaverWnd::Create, Failed to get bounding rect, Error:%u\n"), GetLastError());
    return FALSE;
  }

  //Create the window, use the virtual functions to allow further customisation
  return CreateEx(GetExtendedWindowStyle(bChild, pParentWnd), pszClassName, AfxGetApp()->m_pszAppName, 
                  GetWindowStyle(bChild, pParentWnd), rRect, pParentWnd, 0);
}

#pragma warning(suppress: 26434)
void CScreenSaverWnd::OnSysCommand(_In_ UINT nID, _In_ LPARAM lParam)
{
  if ((nID != SC_SCREENSAVE) && (nID != SC_CLOSE))
    __super::OnSysCommand(nID, lParam);
}

#pragma warning(suppress: 26433 26440)
void CScreenSaverWnd::PostNcDestroy()
{
  //The default for screen saver windows is to allocate them on the heap meaning
  //the default post-cleanup is to 'delete this'.
  //never explicitly call 'delete' on a CScreenSaverWnd, use DestroyWindow instead
  delete this;
}

#pragma warning(suppress: 26434)
BOOL CScreenSaverWnd::OnSetCursor(_In_opt_ CWnd* pWnd, _In_ UINT nHitTest, _In_ UINT message)
{
  if (m_bPreview)
  {
    //Let the base class do its thing
    return __super::OnSetCursor(pWnd, nHitTest, message);
  }
  else
  {
    SetCursor(nullptr);
    return TRUE;
  }
}

#pragma warning(suppress: 26434)
void CScreenSaverWnd::OnActivateApp(_In_ BOOL bActive, _In_ DWORD dwThreadID)
{
  if (!bActive)
  {
    if (!m_bPreview)
      PostMessage(WM_CLOSE);
  }
  else
  {
    //Let the base class do its thing
    __super::OnActivateApp(bActive, dwThreadID);
  }
}

#pragma warning(suppress: 26434)
void CScreenSaverWnd::OnMouseMove(_In_ UINT /*nFlags*/, _In_ CPoint point)
{
  if (m_bGotCursorPos)
  {
    int nDeltaX{point.x - m_LastPos.x};
    if (nDeltaX < 0)
      nDeltaX *= -1;
    int nDeltaY{point.y - m_LastPos.y};
    if (nDeltaY < 0)
      nDeltaY *= -1;
    m_LastPos = point;

    if (((nDeltaX + nDeltaY) > 3) && !m_bPreview)
      PostMessage(WM_CLOSE);
  }
  else
  {
    m_bGotCursorPos = true;
    m_LastPos = point;
  }
}

#pragma warning(suppress: 26434)
void CScreenSaverWnd::OnKeyDown(_In_ UINT /*nChar*/, _In_ UINT /*nRepCnt*/, _In_ UINT /*nFlags*/)
{
  if (!m_bPreview)
    PostMessage(WM_CLOSE);
}

#pragma warning(suppress: 26434)
void CScreenSaverWnd::OnSysKeyDown(_In_ UINT /*nChar*/, _In_ UINT /*nRepCnt*/, _In_ UINT /*nFlags*/)
{
  if (!m_bPreview)
    PostMessage(WM_CLOSE);
}

#pragma warning(suppress: 26434)
void CScreenSaverWnd::OnPaint()
{
  //Display some demo text to remind the developer that they need to override this function in their derived class
  CPaintDC dc{this};
  dc.SetTextColor(RGB(255, 255, 255));
  dc.SetBkMode(TRANSPARENT);
  dc.SetTextAlign(TA_CENTER);
  CRect rClient;
  GetClientRect(&rClient);
  dc.DrawText(_T("You need to override CScreenSaverWnd::OnPaint in your application"), -1, rClient, DT_CENTER | DT_VCENTER | DT_SINGLELINE);
}

#pragma warning(suppress: 26434)
void CScreenSaverWnd::OnActivate(_In_ UINT nState, _In_opt_ CWnd* pWndOther, _In_ BOOL bMinimized)
{
  if (nState == WA_INACTIVE)
  {
    if (!m_bPreview)
      PostMessage(WM_CLOSE);
  }
  else
  {
    //Let the base class do its thing
    __super::OnActivate(nState, pWndOther, bMinimized);
  }
}

#pragma warning(suppress: 26434)
void CScreenSaverWnd::OnLButtonDown(_In_ UINT /*nFlags*/, _In_ CPoint /*point*/)
{
  if (!m_bPreview)
    PostMessage(WM_CLOSE);
}

#pragma warning(suppress: 26434)
void CScreenSaverWnd::OnRButtonDown(_In_ UINT /*nFlags*/, _In_ CPoint /*point*/)
{
  if (!m_bPreview)
    PostMessage(WM_CLOSE);
}

#pragma warning(suppress: 26434)
void CScreenSaverWnd::OnMButtonDown(_In_ UINT /*nFlags*/, _In_ CPoint /*point*/)
{
  if (!m_bPreview)
    PostMessage(WM_CLOSE);
}


#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CScreenSaverApp, CWinApp) //NOLINT(clang-diagnostic-unused-local-typedef, modernize-avoid-c-arrays)
END_MESSAGE_MAP()

CScreenSaverApp::CScreenSaverApp() noexcept : m_pRuntimeScreenSaverWnd{nullptr}
{
}

#pragma warning(suppress: 26440)
void CScreenSaverApp::SetScreenSaverTemplate(_In_ CRuntimeClass* pRuntimeScreenSaverWnd)
{
  //Validate our parameters
#pragma warning(suppress: 26496)
  AFXASSUME(pRuntimeScreenSaverWnd != nullptr);
#pragma warning(suppress: 26486)
  ASSERT(pRuntimeScreenSaverWnd->IsDerivedFrom(RUNTIME_CLASS(CScreenSaverWnd)));

  m_pRuntimeScreenSaverWnd = pRuntimeScreenSaverWnd;
}

void CScreenSaverApp::ShowConfigureDialog(_In_opt_ CWnd* pParentWnd)
{
  CString sMsg{_T("You need to override CScreenSaverApp::ShowConfigureDialog to allow your screen saver to be configured")};
  if (pParentWnd != nullptr)
      MessageBox(pParentWnd->GetSafeHwnd(), sMsg, m_pszAppName, MB_OK);
  else
      MessageBox(nullptr, sMsg, m_pszAppName, MB_OK);
}

#pragma warning(suppress: 26434)
bool CScreenSaverApp::ProcessShellCommand(_In_ CScreenSaverCommandLineInfo& rCmdInfo)
{
  bool bResult{false};
  switch (rCmdInfo.m_nShellCommand)
  {
    case CScreenSaverCommandLineInfo::ShellCommand::ScreenSaverMode:
    {
#pragma warning(suppress: 26496)
      AFXASSUME(m_pRuntimeScreenSaverWnd != nullptr); //Did you forget to call SetScreenSaverTemplate
#pragma warning(suppress: 26466)
      auto pMainWnd{static_cast<CScreenSaverWnd*>(m_pRuntimeScreenSaverWnd->CreateObject())};
      if (pMainWnd == nullptr)
      {
        TRACE(_T("CScreenSaverApp::ProcessShellCommand, Failed to create screen saver object, Error:%u\n"), GetLastError());
        return false;
      }
      ASSERT(pMainWnd->IsKindOf(RUNTIME_CLASS(CScreenSaverWnd)));
      pMainWnd->SetPreviewFlag(false);
      if (!pMainWnd->Create(false, nullptr))
      {
        TRACE(_T("CScreenSaverApp::ProcessShellCommand, Failed to create screen saver window, Error:%u\n"), GetLastError());
        delete pMainWnd;
        return false;
      }
      m_pMainWnd = pMainWnd;
      bResult = true;
      break;
    }
    case CScreenSaverCommandLineInfo::ShellCommand::PreviewMode:
    {
#pragma warning(suppress: 26496)
      AFXASSUME(m_pRuntimeScreenSaverWnd != nullptr); //Did you forget to call SetScreenSaverTemplate
#pragma warning(suppress: 26466)
      auto pMainWnd{static_cast<CScreenSaverWnd*>(m_pRuntimeScreenSaverWnd->CreateObject())};
      if (pMainWnd == nullptr)
      {
        TRACE(_T("CScreenSaverApp::ProcessShellCommand, Failed to create screen saver object, Error:%u\n"), GetLastError());
        return false;
      }
      ASSERT(pMainWnd->IsKindOf(RUNTIME_CLASS(CScreenSaverWnd)));
      pMainWnd->SetPreviewFlag(true);
      CWnd wndParent;
      wndParent.Attach(rCmdInfo.m_hParentWnd);
#pragma warning(suppress: 26486)
      if (!pMainWnd->Create(true, &wndParent))
      {
        TRACE(_T("CScreenSaverApp::ProcessShellCommand, Failed to create screen saver window, Error:%u\n"), GetLastError());
        delete pMainWnd;
        return false;
      }
      m_pMainWnd = pMainWnd;
      wndParent.Detach();
      bResult = true;
      break;
    }
    case CScreenSaverCommandLineInfo::ShellCommand::ShowConfigureDialog:
    {
      if (rCmdInfo.m_hParentWnd)
      {
        CWnd wndParent;
        wndParent.Attach(rCmdInfo.m_hParentWnd);
#pragma warning(suppress: 26486)
        ShowConfigureDialog(&wndParent);
        wndParent.Detach();
      }
      else
        ShowConfigureDialog(CWnd::GetActiveWindow());
      bResult = false;
      break;
    }
    default:
    {
      ASSERT(false);
      break;
    }
  }
  return bResult;
}

#pragma warning(suppress: 26434)
void CScreenSaverApp::ParseCommandLine(_In_ CScreenSaverCommandLineInfo& rCmdInfo) noexcept
{
  for (int i{1}; i<__argc; i++)
  {
#ifdef _UNICODE
#pragma warning(suppress: 26429 26481)
    LPCTSTR pszParam{__wargv[i]};
#else
#pragma warning(suppress: 26429 26481)
    LPCTSTR pszParam{__argv[i]};
#endif
    BOOL bFlag{FALSE};
    const BOOL bLast{(i + 1) == __argc};
#pragma warning(suppress: 26481)
    if (pszParam[0] == _T('-') || pszParam[0] == _T('/'))
    {
        //remove flag specifier
        bFlag = TRUE;
#pragma warning(suppress: 26481)
        ++pszParam;
    }
#pragma warning(suppress: 26486)
    rCmdInfo.ParseParam(pszParam, bFlag, bLast);
  }
}

BOOL CScreenSaverApp::InitInstance()
{
  //Let the base class do its thing
  if (!__super::InitInstance())
    return FALSE;

  //Load up the screen saver description string and setup the app caption
  CString sDescription;
  if (!sDescription.LoadString(1))
  {
    ASSERT(false); //Have you forgot to add a string resource with an ID of 1?
    return FALSE;
  }
#pragma warning(suppress: 26408 26465 26492)
  free(const_cast<LPTSTR>(m_pszAppName));
  m_pszAppName = _tcsdup(sDescription);

  //Parse the command line and act on its results
  CScreenSaverCommandLineInfo cmdInfo;
#pragma warning(suppress: 26486)
  ParseCommandLine(cmdInfo);
  return ProcessShellCommand(cmdInfo);
}
